<script src="{{ base_url }}-/static/sql-formatter-2.3.3.min.js" defer></script>
<script src="{{ base_url }}-/static/codemirror-5.57.0.min.js"></script>
<link rel="stylesheet" href="{{ base_url }}-/static/codemirror-5.57.0.min.css" />
<script src="{{ base_url }}-/static/codemirror-5.57.0-sql.min.js"></script>
<style>
    .CodeMirror { height: auto; min-height: 70px; width: 80%; border: 1px solid #ddd; }
    .CodeMirror-scroll { max-height: 200px; }
</style>
